@@ -2,7 +2,7 @@
 ######################################################################
 ### BER (Basic Encoding Rules) encoding and decoding.
 ######################################################################
-### Copyright (c) 1995-1999, Simon Leinen.
+### Copyright (c) 1995-2000, Simon Leinen.
 ###
 ### This program is free software; you can redistribute it under the
 ### "Artistic License" included in this distribution (file "Artistic").
@@ -22,6 +22,9 @@
 ### Yufang HU <yhu@casc.com>:  Support even larger subids
 ### Mike Mitchell <mcm@unx.sas.com>: New generalized encode_int()
 ### Mike Diehn <mdiehn@mindspring.net>: encode_ip_address()
+### Rik Hoorelbeke <rik.hoorelbeke@pandora.be>: encode_oid() fix
+### Brett T Warden <wardenb@eluminant.com>: pretty UInteger32
+### Bert Driehuis <driehuis@playbeing.org>: Handle SNMPv2 exception codes
 ######################################################################
 
 package BER;
@@ -32,7 +35,7 @@ use strict;
 use vars qw(@ISA @EXPORT $VERSION $pretty_print_timeticks $errmsg);
 use Exporter;
 
-$VERSION = '0.72';
+$VERSION = '0.88';
 
 @ISA = qw(Exporter);
 
@@ -40,6 +43,8 @@ $VERSION = '0.72';
 	     encode_int encode_int_0 encode_null encode_oid
 	     encode_sequence encode_tagged_sequence
 	     encode_string encode_ip_address encode_timeticks
+	     encode_uinteger32 encode_counter32 encode_counter64
+	     encode_gauge32 
 	     decode_sequence decode_by_template
 	     pretty_print pretty_print_timeticks
 	     hex_string hex_string_of_type
@@ -132,6 +137,12 @@ sub snmp_nsap_address_tag	{ 0x05 | application_flag }
 sub snmp_counter64_tag		{ 0x06 | application_flag }
 sub snmp_uinteger32_tag		{ 0x07 | application_flag }
 
+## Error codes (SNMPv2 and later)
+##
+sub snmp_nosuchobject		{ context_flag | 0x00 }
+sub snmp_nosuchinstance		{ context_flag | 0x01 }
+sub snmp_endofmibview		{ context_flag | 0x02 }
+
 #### Encoding
 
 sub encode_header ($$) {
@@ -150,18 +161,44 @@ sub encode_int ($) {
     return encode_intlike ($_[0], int_tag);
 }
 
+sub encode_uinteger32 ($) {
+    return encode_intlike ($_[0], snmp_uinteger32_tag);
+}
+
+sub encode_counter32 ($) {
+    return encode_intlike ($_[0], snmp_counter32_tag);
+}
+
+sub encode_counter64 ($) {
+    return encode_intlike ($_[0], snmp_counter64_tag);
+}
+
+sub encode_gauge32 ($) {
+    return encode_intlike ($_[0], snmp_gauge32_tag);
+}
+
 sub encode_intlike ($$) {
     my ($int, $tag)=@_;
     my ($sign, $val, @vals);
-
     $sign = ($int >= 0) ? 0 : 0xff;
-    for(;;) {
-        $val = $int & 0xff;
-        unshift(@vals, $val);
-        return encode_header ($tag, $#vals + 1).pack ("C*", @vals)
-            if ($int >= -128 && $int < 128);
-        $int -= $sign;
-        $int = int($int / 256);
+    if (ref $int && $int->isa ("Math::BigInt")) {
+	for(;;) {
+	    $val = $int->bmod (256);
+	    unshift(@vals, $val);
+	    return encode_header ($tag, $#vals + 1).pack ("C*", @vals)
+		if ($int >= -128 && $int < 128);
+	    $int = $int - $sign;
+	    $int = $int / 256;
+	}
+    } else {
+	for(;;) {
+	    $val = $int & 0xff;
+	    unshift(@vals, $val);
+	    return encode_header ($tag, $#vals + 1).pack ("C*", @vals)
+		if ($int >= -128 && $int < 128);
+	    $int -= $sign;
+	    $int = int($int / 256);
+	}
     }
 }
 
@@ -176,7 +213,7 @@ sub encode_oid (@) {
     shift @oid if $oid[0] eq '';
 
     return error ("Object ID too short: ", join('.',@oid))
-	if $#oid <= 1;
+	if $#oid < 1;
     ## The first two subids in an Object ID are encoded as a single
     ## byte in BER, according to a funny convention.  This poses
     ## restrictions on the ranges of those subids.  In the past, I
@@ -209,14 +246,7 @@ sub encode_oid (@) {
 			     0x80 | (($subid>>14) & 0x7f),
 			     0x80 | (($subid>>7) & 0x7f),
 			     $subid & 0x7f);
-	} elsif ( ($subid>=268435456) && ($subid<2147483648) ){ #31 bits long subid
-	    $result .= pack ("CCCCC", 
-			     0x80 | (($subid>>28) & 0x0f), #mask the bits beyond 32 
-			     0x80 | (($subid>>21) & 0x7f),
-			     0x80 | (($subid>>14) & 0x7f),
-			     0x80 | (($subid>>7) & 0x7f),
-			     $subid & 0x7f);
-	} elsif ( ($subid>=-2147483648) && ($subid<0) ) { #32 bits long subid
+	} elsif ( ($subid>=268435456) && ($subid<4294967296) ){ #32 bits long subid
 	    $result .= pack ("CCCCC", 
 			     0x80 | (($subid>>28) & 0x0f), #mask the bits beyond 32 
 			     0x80 | (($subid>>21) & 0x7f),
@@ -271,13 +301,15 @@ sub encode_timeticks ($) {
 sub pretty_print ($) {
     my ($packet) = @_;
     my ($type,$rest);
+    return undef unless defined $packet;
     my $result = ord (substr ($packet, 0, 1));
     return pretty_intlike ($packet)
 	if $result == int_tag;
     return pretty_unsignedlike ($packet)
 	if $result == snmp_counter32_tag
 	    || $result == snmp_gauge32_tag
-	    || $result == snmp_counter64_tag;
+	    || $result == snmp_counter64_tag
+	    || $result == snmp_uinteger32_tag;
     return pretty_string ($packet) if $result == octet_string_tag;
     return pretty_oid ($packet) if $result == object_id_tag;
     return ($pretty_print_timeticks
@@ -286,7 +318,10 @@ sub pretty_print ($) {
 	if $result == uptime_tag;
     return pretty_ip_address ($packet) if $result == snmp_ip_address_tag;
     return "(null)" if $result == null_tag;
-    return sprintf("#<unprintable BER type 0x%x>", $result);
+    return error ("Exception code: noSuchObject") if $result == snmp_nosuchobject;
+    return error ("Exception code: noSuchInstance") if $result == snmp_nosuchinstance;
+    return error ("Exception code: endOfMibView") if $result == snmp_endofmibview;
+    return sprintf ("#<unprintable BER type 0x%x>", $result);
 }
 
 sub pretty_using_decoder ($$) {
@@ -595,12 +630,18 @@ sub decode_unsignedlike ($) {
     decode_intlike_s ($_[0], 0);
 }
 
+my $have_math_bigint_p = 0;
+
 sub decode_intlike_s ($$) {
     my ($pdu, $signedp) = @_;
     my ($length,$result);
     $length = ord (substr ($pdu, 1, 1));
     my $ptr = 2;
     $result = unpack ($signedp ? "c" : "C", substr ($pdu, $ptr++, 1));
+    if ($length > 5 || ($length == 5 && $result > 0)) {
+	require 'Math/BigInt.pm' unless $have_math_bigint_p++;
+	$result = new Math::BigInt ($result);
+    }
     while (--$length > 0) {
 	$result *= 256;
 	$result += unpack ("C", substr ($pdu, $ptr++, 1));
