@@ -2,7 +2,7 @@
 ######################################################################
 ### SNMP Request/Response Handling
 ######################################################################
-### Copyright (c) 1995-1999, Simon Leinen.
+### Copyright (c) 1995-2000, Simon Leinen.
 ###
 ### This program is free software; you can redistribute it under the
 ### "Artistic License" included in this distribution (file "Artistic").
@@ -11,8 +11,9 @@
 ### to communicate with SNMP entities.  It has methods to send
 ### requests to and receive responses from an agent.
 ###
-### Currently it has one subclass, SNMPv1_Session, which implements
-### the SNMPv1 protocol.
+### Two instantiable subclasses are defined:
+### SNMPv1_Session implements SNMPv1 (RFC 1157) functionality
+### SNMPv2c_Session implements community-based SNMPv2.
 ######################################################################
 ### Created by:  Simon Leinen  <simon@switch.ch>
 ###
@@ -27,6 +28,9 @@
 ### Alan Nichols <Alan.Nichols@Ebay.Sun.COM>
 ### Mike McCauley <mikem@open.com.au>
 ### Andrew W. Elble <elble@icculus.nsg.nwu.edu>
+### Brett T Warden <wardenb@eluminant.com>: pretty UInteger32
+### Michael Deegan <michael@cnspc18.murdoch.edu.au>
+### Sergio Macedo <macedo@tmp.com.br>
 ######################################################################
 
 package SNMP_Session;		
@@ -40,18 +44,19 @@ use Exporter;
 use vars qw(@ISA $VERSION @EXPORT $errmsg $suppress_warnings);
 use Socket;
 use BER;
+use Carp;
 
-sub map_table ($$$);
+sub map_table ($$$ );
 sub map_table_4 ($$$$);
 sub map_table_start_end ($$$$$$);
 sub index_compare ($$);
 sub oid_diff ($$);
 
-$VERSION = '0.75';
+$VERSION = '0.90';
 
 @ISA = qw(Exporter);
 
-@EXPORT = qw(errmsg suppress_warnings index_compare oid_diff);
+@EXPORT = qw(errmsg suppress_warnings index_compare oid_diff recycle_socket);
 
 my $default_debug = 0;
 
@@ -61,10 +66,12 @@ my $default_debug = 0;
 ###
 my $default_timeout = 2.0;
 
-### Default number of retries for each SNMP request.  If no response
-### is received after TIMEOUT seconds, the request is resent and a new
-### response awaited with a longer timeout (see the documentation on
-### BACKOFF below).
+### Default number of attempts to get a reply for an SNMP request.  If
+### no response is received after TIMEOUT seconds, the request is
+### resent and a new response awaited with a longer timeout (see the
+### documentation on BACKOFF below).  The "retries" value should be at
+### least 1, because the first attempt counts, too (the name "retries"
+### is confusing, sorry for that).
 ###
 my $default_retries = 5;
 
@@ -85,6 +92,12 @@ my $default_backoff = 1.0;
 ###
 my $default_max_repetitions = 12;
 
+### Whether all SNMP_Session objects should share a single UDP socket.
+###
+$SNMP_Session::recycle_socket = 0;
+
+my $the_socket;
+
 $SNMP_Session::errmsg = '';
 $SNMP_Session::suppress_warnings = 0;
 
@@ -109,29 +122,29 @@ sub retries { $_[0]->{retries} }
 sub backoff { $_[0]->{backoff} }
 sub set_timeout {
     my ($session, $timeout) = @_;
-    die "timeout ($timeout) must be a positive number" unless $timeout > 0.0;
+    croak ("timeout ($timeout) must be a positive number") unless $timeout > 0.0;
     $session->{'timeout'} = $timeout;
 }
 sub set_retries {
     my ($session, $retries) = @_;
-    die "retries ($retries) must be a non-negative integer"
+    croak ("retries ($retries) must be a non-negative integer")
 	unless $retries == int ($retries) && $retries >= 0;
     $session->{'retries'} = $retries; 
 }
 sub set_backoff {
     my ($session, $backoff) = @_;
-    die "backoff ($backoff) must be a number >= 1.0"
+    croak ("backoff ($backoff) must be a number >= 1.0")
 	unless $backoff == int ($backoff) && $backoff >= 1.0;
     $session->{'backoff'} = $backoff; 
 }
 
-sub encode_request_3 ($$$@)
-{
+sub encode_request_3 ($$$@) {
     my($this, $reqtype, $encoded_oids_or_pairs, $i1, $i2) = @_;
     my($request);
     local($_);
 
-    ++$this->{request_id};
+    $this->{request_id} = ($this->{request_id} == 0x7fffffff)
+	? -0x80000000 : $this->{request_id}+1;
     foreach $_ (@{$encoded_oids_or_pairs}) {
       if (ref ($_) eq 'ARRAY') {
 	$_ = &encode_sequence ($_->[0], $_->[1])
@@ -151,33 +164,28 @@ sub encode_request_3 ($$$@)
     return $this->wrap_request ($request);
 }
 
-sub encode_get_request
-{
+sub encode_get_request {
     my($this, @oids) = @_;
     return encode_request_3 ($this, get_request, \@oids);
 }
 
-sub encode_getnext_request
-{
+sub encode_getnext_request {
     my($this, @oids) = @_;
     return encode_request_3 ($this, getnext_request, \@oids);
 }
 
-sub encode_getbulk_request
-{
+sub encode_getbulk_request {
     my($this, $non_repeaters, $max_repetitions, @oids) = @_;
     return encode_request_3 ($this, getbulk_request, \@oids,
 			     $non_repeaters, $max_repetitions);
 }
 
-sub encode_set_request
-{
+sub encode_set_request {
     my($this, @encoded_pairs) = @_;
     return encode_request_3 ($this, set_request, \@encoded_pairs);
 }
 
-sub encode_trap_request ($$$$$$@)
-{
+sub encode_trap_request ($$$$$$@) {
     my($this, $ent, $agent, $gen, $spec, $dt, @pairs) = @_;
     my($request);
     local($_);
@@ -197,15 +205,13 @@ sub encode_trap_request ($$$$$$@)
     return $this->wrap_request ($request);
 }
 
-sub encode_v2_trap_request ($@)
-{
+sub encode_v2_trap_request ($@) {
     my($this, @pairs) = @_;
 
     return encode_request_3($this, trap2_request, \@pairs);
 }
 
-sub decode_get_response
-{
+sub decode_get_response {
     my($this, $response) = @_;
     my @rest;
     @{$this->{'unwrapped'}};
@@ -238,8 +244,7 @@ sub decode_trap_request ($$) {
     return ($community, $ent, $agent, $gen, $spec, $dt, $bindings);
 }
 
-sub wait_for_response
-{
+sub wait_for_response {
     my($this) = shift;
     my($timeout) = shift || 10.0;
     my($rin,$win,$ein) = ('','','');
@@ -248,37 +253,32 @@ sub wait_for_response
     select($rout=$rin,$wout=$win,$eout=$ein,$timeout);
 }
 
-sub get_request_response ($@)
-{
+sub get_request_response ($@) {
     my($this, @oids) = @_;
     return $this->request_response_5 ($this->encode_get_request (@oids),
 				      get_response, \@oids, 1);
 }
 
-sub set_request_response ($@)
-{
+sub set_request_response ($@) {
     my($this, @pairs) = @_;
     return $this->request_response_5 ($this->encode_set_request (@pairs),
 				      get_response, \@pairs, 1);
 }
 
-sub getnext_request_response ($@)
-{
+sub getnext_request_response ($@) {
     my($this,@oids) = @_;
     return $this->request_response_5 ($this->encode_getnext_request (@oids),
 				      get_response, \@oids, 1);
 }
 
-sub getbulk_request_response ($$$@)
-{
+sub getbulk_request_response ($$$@) {
     my($this,$non_repeaters,$max_repetitions,@oids) = @_;
     return $this->request_response_5
 	($this->encode_getbulk_request ($non_repeaters,$max_repetitions,@oids),
 	 get_response, \@oids, 1);
 }
 
-sub trap_request_send ($$$$$$@)
-{
+sub trap_request_send ($$$$$$@) {
     my($this, $ent, $agent, $gen, $spec, $dt, @pairs) = @_;
     my($req);
 
@@ -290,8 +290,7 @@ sub trap_request_send ($$$$$$@)
     return 1;
 }
 
-sub v2_trap_request_send ($$$@)
-{
+sub v2_trap_request_send ($$$@) {
     my($this, $trap_oid, $dt, @pairs) = @_;
     my @sysUptime_OID = ( 1,3,6,1,2,1,1,3 );
     my @snmpTrapOID_OID = ( 1,3,6,1,6,3,1,1,4,1 );
@@ -309,8 +308,7 @@ sub v2_trap_request_send ($$$@)
     return 1;
 }
 
-sub request_response_5 ($$$$$)
-{
+sub request_response_5 ($$$$$) {
     my ($this, $req, $response_tag, $oids, $errorp) = @_;
     my $retries = $this->retries;
     my $timeout = $this->timeout;
@@ -320,9 +318,10 @@ sub request_response_5 ($$$$$)
     return undef unless defined $req;
 
     $timeleft = $timeout;
-    $this->send_query ($req)
-	|| return $this->error ("send_query: $!");
     while ($retries > 0) {
+	$this->send_query ($req)
+	    || return $this->error ("send_query: $!");
+      wait_for_response:
 	($nfound, $timeleft) = $this->wait_for_response($timeleft);
 	if ($nfound > 0) {
 	    my($response_length);
@@ -332,8 +331,9 @@ sub request_response_5 ($$$$$)
 	    if ($response_length) {
 		return $response_length;
 	    } elsif (defined ($response_length)) {
+		goto wait_for_response;
 		# A response has been received, but for a different
-		# request ID.
+		# request ID or from a different IP address.
 	    } else {
 		return undef;
 	    }
@@ -342,47 +342,11 @@ sub request_response_5 ($$$$$)
 	    --$retries;
 	    $timeout *= $this->backoff;
 	    $timeleft = $timeout;
-	    $this->send_query ($req)
-		|| return $this->error ("send_query: $!");
 	}
     }
     $this->error ("no response received");
 }
 
-
-sub error_return ($$)
-{
-    my ($this,$message) = @_;
-    $SNMP_Session::errmsg = $message;
-    unless ($SNMP_Session::suppress_warnings) {
-	$message =~ s/^/  /mg;
-	warn ("Error:\n".$message."\n");
-    }
-    return undef;
-}
-
-sub error ($$)
-{
-    my ($this,$message) = @_;
-    my $session = $this->to_string;
-    $SNMP_Session::errmsg = $message."\n".$session;
-    unless ($SNMP_Session::suppress_warnings) {
-	$session =~ s/^/  /mg;
-	$message =~ s/^/  /mg;
-	warn ("SNMP Error:\n".$SNMP_Session::errmsg."\n");
-    }
-    return undef;
-}
-
-sub ber_error ($$)
-{
-  my ($this,$type) = @_;
-  my ($errmsg) = $BER::errmsg;
-
-  $errmsg =~ s/^/  /mg;
-  return $this->error ("$type:\n$errmsg");
-}
-
 sub map_table ($$$) {
     my ($session, $columns, $mapfn) = @_;
     return $session->map_table_4 ($columns, $mapfn,
@@ -486,8 +450,7 @@ sub oid_diff ($$) {
   substr ($full_dotnot, length ($base_dotnot)+1);
 }
 
-sub pretty_address
-{
+sub pretty_address {
     my($addr) = shift;
     my($port,$ipaddr) = unpack_sockaddr_in($addr);
     return sprintf ("[%s].%d",inet_ntoa($ipaddr),$port);
@@ -495,6 +458,37 @@ sub pretty_address
 
 sub version { $VERSION; }
 
+
+sub error_return ($$) {
+    my ($this,$message) = @_;
+    $SNMP_Session::errmsg = $message;
+    unless ($SNMP_Session::suppress_warnings) {
+	$message =~ s/^/  /mg;
+	carp ("Error:\n".$message."\n");
+    }
+    return undef;
+}
+
+sub error ($$) {
+    my ($this,$message) = @_;
+    my $session = $this->to_string;
+    $SNMP_Session::errmsg = $message."\n".$session;
+    unless ($SNMP_Session::suppress_warnings) {
+	$session =~ s/^/  /mg;
+	$message =~ s/^/  /mg;
+	carp ("SNMP Error:\n".$SNMP_Session::errmsg."\n");
+    }
+    return undef;
+}
+
+sub ber_error ($$) {
+  my ($this,$type) = @_;
+  my ($errmsg) = $BER::errmsg;
+
+  $errmsg =~ s/^/  /mg;
+  return $this->error ("$type:\n$errmsg");
+}
+
 package SNMPv1_Session;
 
 use strict qw(vars subs);	# see above
@@ -503,15 +497,17 @@ use SNMP_Session;
 use Socket;
 use BER;
 use IO::Socket;
+use Carp;
 
 @ISA = qw(SNMP_Session);
 
 sub snmp_version { 0 }
 
-sub open
-{
-    my($this,$remote_hostname,$community,$port,
-       $max_pdu_len,$bind_to_port,$max_repetitions) = @_;
+sub open {
+    my($this,
+       $remote_hostname,$community,$port,
+       $max_pdu_len,$local_port,$max_repetitions,
+       $local_hostname) = @_;
     my($remote_addr,$socket);
 
     $community = 'public' unless defined $community;
@@ -519,13 +515,23 @@ sub open
     $max_pdu_len = 8000 unless defined $max_pdu_len;
     $max_repetitions = $default_max_repetitions
 	unless defined $max_repetitions;
-    $remote_addr = inet_aton ($remote_hostname)
-	|| return $this->error_return ("can't resolve \"$remote_hostname\" to IP address");
-    $socket = IO::Socket::INET->new(Proto => 17,
-				    Type => SOCK_DGRAM,
-				    LocalPort => $bind_to_port)
-	|| return $this->error_return ("creating socket: $!");
-    $remote_addr = pack_sockaddr_in ($port, $remote_addr);
+    if (defined $remote_hostname) {
+	$remote_addr = inet_aton ($remote_hostname)
+	    or return $this->error_return ("can't resolve \"$remote_hostname\" to IP address");
+    }
+    if ($SNMP_Session::recycle_socket && defined $the_socket) {
+	$socket = $the_socket;
+    } else {
+	$socket = IO::Socket::INET->new(Proto => 17,
+					Type => SOCK_DGRAM,
+					LocalAddr => $local_hostname,
+					LocalPort => $local_port)
+	    || return $this->error_return ("creating socket: $!");
+	$the_socket = $socket
+	    if $SNMP_Session::recycle_socket;
+    }
+    $remote_addr = pack_sockaddr_in ($port, $remote_addr)
+	if defined $remote_addr;
     bless {
 	   'sock' => $socket,
 	   'sockfileno' => fileno ($socket),
@@ -534,7 +540,8 @@ sub open
 	   'remote_addr' => $remote_addr,
 	   'max_pdu_len' => $max_pdu_len,
 	   'pdu_buffer' => '\0' x $max_pdu_len,
-	   'request_id' => int (rand 0x80000000 + rand 0xffff),
+	   'request_id' => ((int (rand 0x10000) << 16) + int (rand 0x10000))
+	       - 0x80000000,
 	   'timeout' => $default_timeout,
 	   'retries' => $default_retries,
 	   'backoff' => $default_backoff,
@@ -542,13 +549,16 @@ sub open
 	   'error_status' => 0,
 	   'error_index' => 0,
 	   'default_max_repetitions' => $max_repetitions,
+	   'use_getbulk' => 1,
+	   'lenient_source_address_matching' => 1,
+	   'lenient_source_port_matching' => 1,
 	  };
 }
 
 sub open_trap_session (@) {
     my ($this, $port) = @_;
     $port = 162 unless defined $port;
-    return $this->open ("0.0.0.0", "", 161, undef, $port);
+    return $this->open (undef, "", 161, undef, $port);
 }
 
 sub sock { $_[0]->{sock} }
@@ -556,16 +566,22 @@ sub sockfileno { $_[0]->{sockfileno} }
 sub remote_addr { $_[0]->{remote_addr} }
 sub pdu_buffer { $_[0]->{pdu_buffer} }
 sub max_pdu_len { $_[0]->{max_pdu_len} }
-sub default_max_repetitions { $_[0]->{default_max_repetitions} }
+sub default_max_repetitions {
+    defined $_[1]
+	? $_[0]->{default_max_repetitions} = $_[1]
+	    : $_[0]->{default_max_repetitions} }
+sub debug { defined $_[1] ? $_[0]->{debug} = $_[1] : $_[0]->{debug} }
 
-sub close
-{
+sub close {
     my($this) = shift;
-    close ($this->sock) || $this->error ("close: $!");
+    ## Avoid closing the socket if it may be shared with other session
+    ## objects.
+    if (! defined $the_socket || $this->sock ne $the_socket) {
+	close ($this->sock) || $this->error ("close: $!");
+    }
 }
 
-sub wrap_request
-{
+sub wrap_request {
     my($this) = shift;
     my($request) = shift;
 
@@ -582,8 +598,7 @@ my @error_status_code = qw(noError tooBig noSuchName badValue readOnly
 			   commitFailed undoFailed authorizationError
 			   notWritable inconsistentName);
 
-sub unwrap_response_5b
-{
+sub unwrap_response_5b {
     my ($this,$response,$tag,$oids,$errorp) = @_;
     my ($community,$request_id,@rest,$snmpver);
 
@@ -621,14 +636,37 @@ sub unwrap_response_5b
     ($community, $request_id, @rest);
 }
 
-sub send_query ($$)
-{
+sub send_query ($$) {
     my ($this,$query) = @_;
     send ($this->sock,$query,0,$this->remote_addr);
 }
 
-sub receive_response_3
-{
+## Compare two sockaddr_in structures for equality.  This is used when
+## matching incoming responses with outstanding requests.  Previous
+## versions of the code simply did a bytewise comparison ("eq") of the
+## two sockaddr_in structures, but this didn't work on some systems
+## where sockaddr_in contains other elements than just the IP address
+## and port number, notably FreeBSD.
+##
+## We allow for varying degrees of leniency when checking the source
+## address.  By default we now ignore it altogether, because there are
+## agents that don't respond from UDP port 161, and there are agents
+## that don't respond from the IP address the query had been sent to.
+##
+sub sa_equal_p ($$$) {
+    my ($this, $sa1, $sa2) = @_;
+    my ($p1, $a1) = sockaddr_in ($sa1);
+    my ($p2, $a2) = sockaddr_in ($sa2);
+    if (! $this->{'lenient_source_address_matching'}) {
+	return 0 if $a1 ne $a2;
+    }
+    if (! $this->{'lenient_source_port_matching'}) {
+	return 0 if $p1 != $p2;
+    }
+    return 1;
+}
+
+sub receive_response_3 {
     my ($this, $response_tag, $oids, $errorp) = @_;
     my ($remote_addr);
     $remote_addr = recv ($this->sock,$this->{'pdu_buffer'},$this->max_pdu_len,0);
@@ -643,10 +681,15 @@ sub receive_response_3
     ## request to.  If this is not the case, we should probably ignore
     ## it, as it may relate to another request.
     ##
-    if ($this->{'debug'} && $remote_addr ne $this->{'remote_addr'}) {
-	warn "Response came from ".&SNMP_Session::pretty_address($remote_addr)
-	    .", not ".&SNMP_Session::pretty_address($this->{'remote_addr'})
-		unless $SNMP_Session::suppress_warnings;
+    if (defined $this->{'remote_addr'}) {
+	if (! $this->sa_equal_p ($remote_addr, $this->{'remote_addr'})) {
+	    if ($this->{'debug'} && !$SNMP_Session::recycle_socket) {
+		carp ("Response came from ".&SNMP_Session::pretty_address($remote_addr)
+		      .", not ".&SNMP_Session::pretty_address($this->{'remote_addr'}))
+			unless $SNMP_Session::suppress_warnings;
+	    }
+	    return 0;
+	}
     }
     $this->{'last_sender_addr'} = $remote_addr;
     my ($response_community, $response_id, @unwrapped)
@@ -655,10 +698,10 @@ sub receive_response_3
     if ($response_community ne $this->{community}
         || $response_id ne $this->{request_id}) {
 	if ($this->{'debug'}) {
-	    warn "$response_community != $this->{community}"
+	    carp ("$response_community != $this->{community}")
 		unless $SNMP_Session::suppress_warnings
 		    || $response_community eq $this->{community};
-	    warn "$response_id != $this->{request_id}"
+	    carp ("$response_id != $this->{request_id}")
 		unless $SNMP_Session::suppress_warnings
 		    || $response_id == $this->{request_id};
 	}
@@ -672,8 +715,7 @@ sub receive_response_3
     return length $this->pdu_buffer;
 }
 
-sub receive_trap
-{
+sub receive_trap {
     my ($this) = @_;
     my ($remote_addr, $iaddr, $port, $trap);
     $remote_addr = recv ($this->sock,$this->{'pdu_buffer'},$this->max_pdu_len,0);
@@ -683,21 +725,23 @@ sub receive_trap
     return ($trap, $iaddr, $port);
 }
 
-sub describe
-{
+sub describe {
     my($this) = shift;
     print $this->to_string (),"\n";
 }
 
-sub to_string
-{
+sub to_string {
     my($this) = shift;
     my ($class,$prefix);
 
     $class = ref($this);
     $prefix = ' ' x (length ($class) + 2);
-    ($class." (remote host: \"".$this->{remote_hostname}
-     ."\" ".&SNMP_Session::pretty_address ($this->remote_addr)."\n"
+    ($class
+     .(defined $this->{remote_hostname}
+       ? " (remote host: \"".$this->{remote_hostname}."\""
+       ." ".&SNMP_Session::pretty_address ($this->remote_addr).")"
+       : " (no remote host specified)")
+     ."\n"
      .$prefix."  community: \"".$this->{'community'}."\"\n"
      .$prefix." request ID: ".$this->{'request_id'}."\n"
      .$prefix."PDU bufsize: ".$this->{'max_pdu_len'}." bytes\n"
@@ -712,8 +756,7 @@ sub to_string
 ### SNMP Agent support
 ### contributed by Mike McCauley <mikem@open.com.au>
 ###
-sub receive_request
-{
+sub receive_request {
     my ($this) = @_;
     my ($remote_addr, $iaddr, $port, $request);
 
@@ -725,8 +768,7 @@ sub receive_request
     return ($request, $iaddr, $port);
 }
 
-sub decode_request
-{
+sub decode_request {
     my ($this, $request) = @_;
     my ($snmp_version, $community, $requestid, $errorstatus, $errorindex, $bindings);
 
@@ -764,13 +806,13 @@ use strict qw(vars subs);	# see above
 use vars qw(@ISA);
 use SNMP_Session;
 use BER;
+use Carp;
 
 @ISA = qw(SNMPv1_Session);
 
 sub snmp_version { 1 }
 
-sub open
-{
+sub open {
     my $session = SNMPv1_Session::open (@_);
     return bless $session;
 }
@@ -781,10 +823,16 @@ sub map_table_start_end ($$$$$$) {
     my @encoded_oids;
     my $call_counter = 0;
     my $base_index = $start;
+    my $ncols = @{$columns};
+    my @collected_values = ();
+
+    if (! $session->{'use_getbulk'}) {
+	return SNMP_Session::map_table_start_end ($session, $columns, $mapfn, $start, $end, $max_repetitions);
+    }
     $max_repetitions = $session->default_max_repetitions
 	unless defined $max_repetitions;
 
-    do {
+    for (;;) {
 	foreach (@encoded_oids = @{$columns}) {
 	    $_=encode_oid (@{$_},split '\.',$base_index)
 		|| return $session->ber_error ("encoding OID $base_index");
@@ -793,57 +841,72 @@ sub map_table_start_end ($$$$$$) {
 						@encoded_oids)) {
 	    my $response = $session->pdu_buffer;
 	    my ($bindings) = $session->decode_get_response ($response);
-	    my $smallest_index = undef;
-	    my @collected_values = ();
+	    my @colstack = ();
+	    my $k = 0;
+	    my $j;
+
+	    my $min_index = undef;
 
 	    my @bases = @{$columns};
 	    my $n_bindings = 0;
+	    my $binding;
+
 	    while ($bindings ne '') {
-		my ($binding, $oid, $value);
-		unless (defined $bases[0]) {
-		    @bases = @{$columns};
-		    (++$call_counter,
-		     &$mapfn ($smallest_index, @collected_values))
-			if defined $smallest_index;
-		    $base_index = $smallest_index;
-		    $smallest_index = undef;
-		    @collected_values = ();
-		}
-		my $base = shift @bases;
-
 		($binding, $bindings) = decode_sequence ($bindings);
-		($oid, $value) = decode_by_template ($binding, "%O%@");
+		my ($oid, $value) = decode_by_template ($binding, "%O%@");
 
-		my $out_index;
-
-		++$n_bindings;
-		$out_index = SNMP_Session::oid_diff ($base, $oid);
-		my $cmp;
-		if (!defined $smallest_index
-		    || ($cmp = SNMP_Session::index_compare
-			($out_index,$smallest_index)) == -1) {
-		    $smallest_index = $out_index;
-		    grep ($_=undef, @collected_values);
-		    push @collected_values, $value;
-		} elsif ($cmp == 1) {
-		    push @collected_values, undef;
-		} else {
-		    push @collected_values, $value;
-		}
+		push @{$colstack[$k]}, [$oid, $value];
+		++$k; $k = 0 if $k >= $ncols;
 	    }
-	    @bases = @{$columns};
-	    (++$call_counter,
-	     &$mapfn ($smallest_index, @collected_values))
-		if defined $smallest_index;
-	    $base_index = $smallest_index;
-	    $smallest_index = undef;
-	    @collected_values = ();
+
+	    my $last_min_index = undef;
+	  walk_rows_from_pdu:
+	    for (;;) {
+		my $min_index = undef;
+
+		for ($k = 0; $k < $ncols; ++$k) {
+		    $collected_values[$k] = undef;
+		    my $pair = $colstack[$k]->[0];
+		    unless (defined $pair) {
+			$min_index = undef;
+			last walk_rows_from_pdu;
+		    }
+		    my $this_index
+			= SNMP_Session::oid_diff ($columns->[$k], $pair->[0]);
+		    if (defined $this_index) {
+			my $cmp
+			    = !defined $min_index
+				? -1
+				    : SNMP_Session::index_compare
+					($this_index, $min_index);
+			if ($cmp == -1) {
+			    for ($j = 0; $j < $k; ++$j) {
+				unshift (@{$colstack[$j]},
+					 [$min_index,
+					  $collected_values[$j]]);
+				$collected_values[$j] = undef;
+			    }
+			    $min_index = $this_index;
+			}
+			if ($cmp <= 0) {
+			    $collected_values[$k] = $pair->[1];
+			    shift @{$colstack[$k]};
+			}
+		    }
+		}
+		last unless defined $min_index;
+		last if defined $end && index_compare ($min_index, $end) >= 0;
+		&$mapfn ($min_index, @collected_values);
+		++$call_counter;
+		$last_min_index = $min_index;
+	    }
+	    $base_index = $last_min_index;
 	} else {
 	    return undef;
 	}
+	last unless (defined $base_index
+		     && (!defined $end || index_compare ($base_index, $end) < 0));
     }
-    while (defined $base_index
-	   && (!defined $end || index_compare ($base_index, $end) < 0));
     $call_counter;
 }
 
