@@ -1,20 +1,46 @@
+### - *- mode: Perl -*-
+######################################################################
+### SNMP_util -- SNMP utilities using SNMP_Session.pm and BER.pm
+######################################################################
+### Copyright (c) 1998-2001, Mike Mitchell.
+###
+### This program is free software; you can redistribute it under the
+### "Artistic License" included in this distribution (file "Artistic").
+######################################################################
+### Created by:  Mike Mitchell   <Mike.Mitchell@sas.com>
+###
+### Contributions and fixes by:
+###
+### Tobias Oetiker <oetiker@ee.ethz.ch>:  Basic layout
+### Simon Leinen <simon@switch.ch>: SNMP_session.pm/BER.pm
+### Jeff Allen <jeff.allen@acm.org>: length() of undefined value
+### Johannes Demel <demel@zid.tuwien.ac.at>: MIB file parse problem
+### Simon Leinen <simon@switch.ch>: more OIDs from Interface MIB
+### Jacques Supcik <supcik@ip-plus.net>: Specify local IP, port
+### Tobias Oetiker <oetiker@ee.ethz.ch>: HASH as first OID to set SNMP options
+### Simon Leinen <simon@switch.ch>: 'undefined port' bug
+### Daniel McDonald <dmcdonald@digicontech.com>: request for getbulk support
+### Laurent Girod <girod.laurent@pmintl.ch>: code for snmpwalkhash
+######################################################################
+
 package SNMP_util;
 
-require 5.002;
+require 5.004;
 
 use strict;
 use vars qw(@ISA @EXPORT $VERSION);
 use Exporter;
+use Carp;
 
-use BER "0.58";
-use SNMP_Session "0.59";
+use BER "0.82";
+use SNMP_Session "0.83";
 use Socket;
 
-$VERSION = '0.72';
+$VERSION = '0.90';
 
 @ISA = qw(Exporter);
 
-@EXPORT = qw(snmpget snmpgetnext snmpwalk snmpset snmptrap snmpmapOID snmpMIB_to_OID snmpLoad_OID_Cache snmpQueue_MIB_File);
+@EXPORT = qw(snmpget snmpgetnext snmpwalk snmpset snmptrap snmpgetbulk snmpmaptable snmpwalkhash snmpmapOID snmpMIB_to_OID snmpLoad_OID_Cache snmpQueue_MIB_File);
 
 # The OID numbers from RFC1213 (MIB-II) and RFC1315 (Frame Relay)
 # are pre-loaded below.
@@ -29,13 +55,13 @@ $VERSION = '0.72';
     'mib-2' => '1.3.6.1.2.1',
     'system' => '1.3.6.1.2.1.1',
     'sysDescr' => '1.3.6.1.2.1.1.1.0',
-    'sysObjectID' => '1.3.6.1.2.1.1.2',
+    'sysObjectID' => '1.3.6.1.2.1.1.2.0',
     'sysUpTime' => '1.3.6.1.2.1.1.3.0',
     'sysUptime' => '1.3.6.1.2.1.1.3.0',
     'sysContact' => '1.3.6.1.2.1.1.4.0',
     'sysName' => '1.3.6.1.2.1.1.5.0',
     'sysLocation' => '1.3.6.1.2.1.1.6.0',
-    'sysServices' => '1.3.6.1.2.1.1.7',
+    'sysServices' => '1.3.6.1.2.1.1.7.0',
     'interfaces' => '1.3.6.1.2.1.2',
     'ifNumber' => '1.3.6.1.2.1.2.1.0',
     'ifTable' => '1.3.6.1.2.1.2.2',
@@ -269,28 +295,57 @@ $VERSION = '0.72';
     'snmpInASNParseErrs' => '1.3.6.1.2.1.11.6',
     'snmpInTooBigs' => '1.3.6.1.2.1.11.8',
     'snmpInNoSuchNames' => '1.3.6.1.2.1.11.9',
+    'ifName' => '1.3.6.1.2.1.31.1.1.1.1',
+    'ifInMulticastPkts' => '1.3.6.1.2.1.31.1.1.1.2',
+    'ifInBroadcastPkts' => '1.3.6.1.2.1.31.1.1.1.3',
+    'ifOutMulticastPkts' => '1.3.6.1.2.1.31.1.1.1.4',
+    'ifOutBroadcastPkts' => '1.3.6.1.2.1.31.1.1.1.5',
+    'ifHCInOctets' => '1.3.6.1.2.1.31.1.1.1.6',
+    'ifHCInUcastPkts' => '1.3.6.1.2.1.31.1.1.1.7',
+    'ifHCInMulticastPkts' => '1.3.6.1.2.1.31.1.1.1.8',
+    'ifHCInBroadcastPkts' => '1.3.6.1.2.1.31.1.1.1.9',
+    'ifHCOutOctets' => '1.3.6.1.2.1.31.1.1.1.10',
+    'ifHCOutUcastPkts' => '1.3.6.1.2.1.31.1.1.1.11',
+    'ifHCOutMulticastPkts' => '1.3.6.1.2.1.31.1.1.1.12',
+    'ifHCOutBroadcastPkts' => '1.3.6.1.2.1.31.1.1.1.13',
+    'ifLinkUpDownTrapEnable' => '1.3.6.1.2.1.31.1.1.1.14',
+    'ifHighSpeed' => '1.3.6.1.2.1.31.1.1.1.15',
+    'ifPromiscuousMode' => '1.3.6.1.2.1.31.1.1.1.16',
+    'ifConnectorPresent' => '1.3.6.1.2.1.31.1.1.1.17',
+    'ifAlias' => '1.3.6.1.2.1.31.1.1.1.18',
+    'ifCounterDiscontinuityTime' => '1.3.6.1.2.1.31.1.1.1.19',
     'experimental' => '1.3.6.1.3',
     'private' => '1.3.6.1.4',
     'enterprises' => '1.3.6.1.4.1',
   );
 
+# GIL
+my %revOIDS = ();	# Reversed %SNMP_util::OIDS hash
+my $RevNeeded = 1;
+
 my $agent_start_time = time;
 
 undef $SNMP_util::Host;
 undef $SNMP_util::Session;
+undef $SNMP_util::Version;
+undef $SNMP_util::LHost;
 $SNMP_util::Debug = 0;
 $SNMP_util::CacheFile = "OID_cache.txt";
 $SNMP_util::CacheLoaded = 0;
 
-srand(time|$$);
+srand(time + $$);
 
 ### Prototypes
-sub snmpget (@);
-sub snmpgetnext (@);
-sub snmpopen (@);
-sub snmpwalk (@);
-sub snmpset (@);
-sub snmptrap (@);
+sub snmpget ($@);
+sub snmpgetnext ($@);
+sub snmpopen ($$$);
+sub snmpwalk ($@);
+sub snmpwalk_flg ($$@);
+sub snmpset ($@);
+sub snmptrap ($$$$$@);
+sub snmpgetbulk ($$$@);
+sub snmpmaptable ($$@);
+sub snmpwalkhash ($$@);
 sub toOID (@);
 sub snmpmapOID (@);
 sub snmpMIB_to_OID ($);
@@ -304,35 +359,71 @@ sub version () { $VERSION; }
 #
 # Start an snmp session
 #
-sub snmpopen (@) {
-  my($host) = @_;
-  my($nhost, $port, $community);
-  my($timeout, $retries, $backoff);
+sub snmpopen ($$$) {
+  my($host, $type, $vars) = @_;
+  my($nhost, $port, $community, $lhost, $lport, $nlhost);
+  my($timeout, $retries, $backoff, $version);
 
+  $type = 0 if (!defined($type));
   $community = "public";
-  $port = 161;
+  $nlhost = "";
 
   ($community, $host) = ($1, $2) if ($host =~ /^(.*)@([^@]+)$/);
-  ($host, $port, $timeout, $retries, $backoff) = split(':', $host, 5)
+  ($host, $port, $timeout, $retries, $backoff, $version) = split(':', $host, 6)
     if ($host =~ /:/);
-  $port = 161 if (defined($port) && (length($port) <= 0));
-  $nhost = "$community\@$host:$port";
+  $version = '1' unless defined $version;
+  if (defined($port) && ($port =~ /^([^!]*)!(.*)$/)) {
+    ($port, $lhost) = ($1, $2);
+    $nlhost = $lhost;
+    ($lhost, $lport) = ($1, $2) if ($lhost =~ /^(.*)!(.*)$/);
+    undef($lhost) if (defined($lhost) && (length($lhost) <= 0));
+    undef($lport) if (defined($lport) && (length($lport) <= 0));
+  }
+  undef($port) if (defined($port) && length($port) <= 0);
+  $port = 162 if ($type == 1 && !defined($port));
+  $nhost = "$community\@$host";
+  $nhost .= ":" . $port if (defined($port));
 
   if ((!defined($SNMP_util::Session))
-    || ($SNMP_util::Host ne $nhost))
-  {
-    if (defined($SNMP_util::Session))
-    {
+    || ($SNMP_util::Host ne $nhost)
+    || ($SNMP_util::Version ne $version)
+    || ($SNMP_util::LHost ne $nlhost)) {
+    if (defined($SNMP_util::Session)) {
       $SNMP_util::Session->close();    
       undef $SNMP_util::Session;
       undef $SNMP_util::Host;
+      undef $SNMP_util::Version;
+      undef $SNMP_util::LHost;
     }
-    $SNMP_util::Session = SNMP_Session->open($host, $community, $port);
-    $SNMP_util::Host = $nhost if defined($SNMP_util::Session);
+    $SNMP_util::Session = ($version =~ /^2c?$/i)
+      ? SNMPv2c_Session->open($host, $community, $port, undef,
+				$lport, undef, $lhost)
+      : SNMP_Session->open($host, $community, $port, undef,
+				$lport, undef, $lhost);
+    ($SNMP_util::Host = $nhost, $SNMP_util::Version = $version,
+      $SNMP_util::LHost = $nlhost) if defined($SNMP_util::Session);
   }
 
-  if (defined($SNMP_util::Session))
-  {
+  if (defined($SNMP_util::Session)) {
+    if (ref $vars->[0] eq 'HASH') {
+      my $opts = shift @$vars;
+      foreach $type (keys %$opts) {
+	if (exists $SNMP_util::Session->{$type}) {
+	  if ($type eq 'timeout') {
+	    $SNMP_util::Session->set_timeout($opts->{$type});
+	  } elsif ($type eq 'retries') {
+	    $SNMP_util::Session->set_retries($opts->{$type});
+	  } elsif ($type eq 'backoff') {
+	    $SNMP_util::Session->set_backoff($opts->{$type});
+	  } else {
+	    $SNMP_util::Session->{$type} = $opts->{$type};
+	  }
+	} else {
+	  carp "SNMPopen Unknown SNMP Option Key '$type'\n"
+	    unless ($SNMP_Session::suppress_warnings > 1);
+	}
+      }
+    }
     $SNMP_util::Session->set_timeout($timeout)
       if (defined($timeout) && (length($timeout) > 0));
     $SNMP_util::Session->set_retries($retries)
@@ -347,14 +438,14 @@ sub snmpopen (@) {
 #
 # A restricted snmpget.
 #
-sub snmpget (@) {
+sub snmpget ($@) {
   my($host, @vars) = @_;
   my(@enoid, $var, $response, $bindings, $binding, $value, $oid, @retvals);
   my $session;
 
-  $session = &snmpopen($host);
+  $session = &snmpopen($host, 0, \@vars);
   if (!defined($session)) {
-    warn "SNMPGET Problem for $host\n"
+    carp "SNMPGET Problem for $host\n"
       unless ($SNMP_Session::suppress_warnings > 1);
     return undef;
   }
@@ -374,7 +465,7 @@ sub snmpget (@) {
     return(@retvals);
   }
   $var = join(' ', @vars);
-  warn "SNMPGET Problem for $var on $host\n"
+  carp "SNMPGET Problem for $var on $host\n"
     unless ($SNMP_Session::suppress_warnings > 1);
   return undef;
 }
@@ -382,46 +473,43 @@ sub snmpget (@) {
 #
 # A restricted snmpgetnext.
 #
-sub snmpgetnext (@) {
+sub snmpgetnext ($@) {
   my($host, @vars) = @_;
   my(@enoid, $var, $response, $bindings, $binding);
   my($value, $upoid, $oid, @retvals);
-  my($noid, $ok);
+  my($noid);
   my $session;
 
-  $session = &snmpopen($host);
+  $session = &snmpopen($host, 0, \@vars);
   if (!defined($session)) {
-    warn "SNMPGETNEXT Problem for $host\n"
+    carp "SNMPGETNEXT Problem for $host\n"
       unless ($SNMP_Session::suppress_warnings > 1);
     return undef;
   }
 
   @enoid = &toOID(@vars);
+  return undef unless defined $enoid[0];
 
   undef @vars;
   undef @retvals;
-  foreach $noid (@enoid)
-  {
+  foreach $noid (@enoid) {
     $upoid = pretty_print($noid);
     push(@vars, $upoid);
   }
-  if ($session->getnext_request_response(@enoid))
-  {
+  if ($session->getnext_request_response(@enoid)) {
     $response = $session->pdu_buffer;
     ($bindings) = $session->decode_get_response($response);
     while ($bindings) {
       ($binding, $bindings) = decode_sequence($bindings);
       ($oid, $value) = decode_by_template($binding, "%O%@");
       my $tempo = pretty_print($oid);
-	my $tempv = pretty_print($value);
-	push @retvals, "$tempo:$tempv";
-      }
+      my $tempv = pretty_print($value);
+      push @retvals, "$tempo:$tempv";
+    }
     return (@retvals);
-  }
-  else
-  {
+  } else {
     $var = join(' ', @vars);
-    warn "SNMPGETNEXT Problem for $var on $host\n"
+    carp "SNMPGETNEXT Problem for $var on $host\n"
       unless ($SNMP_Session::suppress_warnings > 1);
     return undef;
   }
@@ -430,31 +518,59 @@ sub snmpgetnext (@) {
 #
 # A restricted snmpwalk.
 #
-sub snmpwalk (@) {
+sub snmpwalk ($@) {
   my($host, @vars) = @_;
+  return(&snmpwalk_flg($host, undef, @vars));
+}
+
+#
+# Walk the MIB, putting everything you find into hashes.
+#
+sub snmpwalkhash($$@) {
+#  my($host, $hash_sub, @vars) = @_;
+  return(&snmpwalk_flg( @_ ));
+}
+
+sub snmpwalk_flg ($$@) {
+  my($host, $hash_sub, @vars) = @_;
   my(@enoid, $var, $response, $bindings, $binding);
   my($value, $upoid, $oid, @retvals);
   my($got, @nnoid, $noid, $ok);
   my $session;
-
-  $session = &snmpopen($host);
+  my(%soid);
+  my(%done, %rethash);
+	
+  $session = &snmpopen($host, 0, \@vars);
   if (!defined($session)) {
-    warn "SNMPWALK Problem for $host\n"
+    carp "SNMPWALK Problem for $host\n"
       unless ($SNMP_Session::suppress_warnings > 1);
     return undef;
   }
 
   @enoid = toOID(@vars);
+  return undef unless defined $enoid[0];
+
+  # GIL
+  #
+  # Create/Refresh a reversed hash with oid -> name
+  #
+  if (defined($hash_sub) && $RevNeeded) {
+    %revOIDS = reverse %SNMP_util::OIDS;
+    $RevNeeded = 0;
+  }
 
   $got = 0;
   @nnoid = @enoid;
   undef @vars;
-  foreach $noid (@enoid)
-  {
+  foreach $noid (@enoid) {
     $upoid = pretty_print($noid);
     push(@vars, $upoid);
   }
-  while($session->getnext_request_response(@nnoid))
+  while(($SNMP_util::Version ne '1' && $session->{'use_getbulk'})
+    ? $session->getbulk_request_response(0,
+					  $session->default_max_repetitions(),
+					  @nnoid)
+    : $session->getnext_request_response(@nnoid))
   {
     $got = 1;
     $response = $session->pdu_buffer;
@@ -465,35 +581,50 @@ sub snmpwalk (@) {
       ($oid, $value) = decode_by_template($binding, "%O%@");
       $ok = 0;
       my $tempo = pretty_print($oid);
-      foreach $noid (@vars)
-      {
-	if ($tempo =~ /^$noid\./ || $tempo eq $noid )
-	{
+      foreach $noid (@vars) {
+	if ($tempo =~ /^$noid\./ || $tempo eq $noid ) {
 	  $ok = 1;
 	  $upoid = $noid;
 	  last;
 	}
       }
-      if ($ok)
-      {
+      if ($ok) {
 	my $tmp = encode_oid_with_errmsg ($tempo);
 	return undef unless defined $tmp;
+	next if (exists($done{$tmp}));	# GIL
 	push @nnoid, $tmp;
 	my $tempv = pretty_print($value);
-	$tempo=~s/^$upoid\.//;
-	push @retvals, "$tempo:$tempv";
+	if (defined($hash_sub)) {
+	  #
+	  # extract name of the oid, if possible, the rest becomes the instance
+	  #
+	  my $inst;
+	  while (!exists($revOIDS{$tempo}) && length($tempo)) {
+	    $tempo =~ s/(\.\d+?)$//;
+	    $inst = $1.$inst;
+	  }	
+	  #
+	  # call hash_sub
+	  #
+	  &$hash_sub(\%rethash, $host, $revOIDS{$tempo}, $tempo, $inst, $tempv);
+	} else {
+	  $tempo=~s/^$upoid\.//;
+	  push @retvals, "$tempo:$tempv";
+	}
+	$done{$tmp} = 1;	# GIL
       }
     }
     last if ($#nnoid < 0);
   }
-  if ($got)
-  {
-    return (@retvals);
-  }
-  else
-  {
+  if ($got) {
+    if (defined($hash_sub)) {
+    	return (%rethash)
+    } else {
+    	return (@retvals);
+    }
+  } else {
     $var = join(' ', @vars);
-    warn "SNMPWALK Problem for $var on $host\n"
+    carp "SNMPWALK Problem for $var on $host\n"
       unless ($SNMP_Session::suppress_warnings > 1);
     return undef;
   }
@@ -502,187 +633,237 @@ sub snmpwalk (@) {
 #
 # A restricted snmpset.
 #
-sub snmpset(@) {
-    my($host, @vars) = @_;
-    my(@enoid, $response, $bindings, $binding);
-    my($oid, @retvals, $type, $value);
-    my $session;
+sub snmpset($@) {
+  my($host, @vars) = @_;
+  my(@enoid, $response, $bindings, $binding);
+  my($oid, @retvals, $type, $value);
+  my $session;
 
-    $session = &snmpopen($host);
-    if (!defined($session))
-    {
-	warn "SNMPSET Problem for $host\n"
-	    unless ($SNMP_Session::suppress_warnings > 1);
-	return undef;
-    }
-
-    while(@vars)
-    {
-	($oid) = toOID((shift @vars));
-	$type  = shift @vars;
-	$value = shift @vars;
-	if ($type =~ /string/i)
-	{
-	    $value = encode_string($value);
-	    push @enoid, [$oid,$value];
-	}
-	elsif ($type =~ /ipaddr/i)
-	{
-	    $value = encode_ip_address($value);
-	    push @enoid, [$oid,$value];
-	}
-	elsif ($type =~ /int/i)
-	{
-	    $value = encode_int($value);
-	    push @enoid, [$oid,$value];
-	}
-	elsif ($type =~ /oid/i)
-	{
-	    my $tmp = encode_oid_with_errmsg($value);
-	    return undef unless defined $tmp;
-	    push @enoid, [$oid,$tmp];
-	}
-	else
-	{
-	    warn "unknown SNMP type: $type\n"
-		unless ($SNMP_Session::suppress_warnings > 1);
-	    return undef;
-	}
-    }
-    if ($session->set_request_response(@enoid))
-    {
-	$response = $session->pdu_buffer;
-	($bindings) = $session->decode_get_response($response);
-	while ($bindings)
-	{
-	    ($binding, $bindings) = decode_sequence($bindings);
-	    ($oid, $value) = decode_by_template($binding, "%O%@");
-	    my $tempo = pretty_print($value);
-	    push @retvals, $tempo;
-	}
-	return (@retvals);
-    }
+  $session = &snmpopen($host, 0, \@vars);
+  if (!defined($session)) {
+    carp "SNMPSET Problem for $host\n"
+      unless ($SNMP_Session::suppress_warnings > 1);
     return undef;
+  }
+
+  while(@vars) {
+    ($oid) = toOID((shift @vars));
+    $type  = shift @vars;
+    $value = shift @vars;
+    if ($type =~ /string/i) {
+      $value = encode_string($value);
+      push @enoid, [$oid,$value];
+    } elsif ($type =~ /ipaddr/i) {
+      $value = encode_ip_address($value);
+      push @enoid, [$oid,$value];
+    } elsif ($type =~ /int/i) {
+      $value = encode_int($value);
+      push @enoid, [$oid,$value];
+    } elsif ($type =~ /oid/i) {
+      my $tmp = encode_oid_with_errmsg($value);
+      return undef unless defined $tmp;
+      push @enoid, [$oid,$tmp];
+    } else {
+      carp "unknown SNMP type: $type\n"
+	unless ($SNMP_Session::suppress_warnings > 1);
+      return undef;
+    }
+  }
+  return undef unless defined $enoid[0];
+  if ($session->set_request_response(@enoid)) {
+    $response = $session->pdu_buffer;
+    ($bindings) = $session->decode_get_response($response);
+    while ($bindings) {
+      ($binding, $bindings) = decode_sequence($bindings);
+      ($oid, $value) = decode_by_template($binding, "%O%@");
+      my $tempo = pretty_print($value);
+      push @retvals, $tempo;
+    }
+    return (@retvals);
+  }
+  return undef;
 }
 
 #
 # Send an SNMP trap
 #
-sub snmptrap(@) {
-    my($host, $ent, $agent, $gen, $spec, @vars) = @_;
-    my($oid, @retvals, $type, $value);
-    my(@enoid);
-    my $session;
+sub snmptrap($$$$$@) {
+  my($host, $ent, $agent, $gen, $spec, @vars) = @_;
+  my($oid, @retvals, $type, $value);
+  my(@enoid);
+  my $session;
 
-    $host = $host . ':162' if !($host =~ /:/);
-    $session = &snmpopen($host);
-    if (!defined($session))
-    {
-	warn "SNMPTRAP Problem for $host\n"
-	    unless ($SNMP_Session::suppress_warnings > 1);
-	return undef;
-    }
+  $session = &snmpopen($host, 1, \@vars);
+  if (!defined($session)) {
+    carp "SNMPTRAP Problem for $host\n"
+      unless ($SNMP_Session::suppress_warnings > 1);
+    return undef;
+  }
 
-    if ($agent =~ /^\d+\.\d+\.\d+\.\d+(.*)/ )
-    {
-	$agent = pack("C*", split /\./, $agent);
+  if ($agent =~ /^\d+\.\d+\.\d+\.\d+(.*)/ ) {
+    $agent = pack("C*", split /\./, $agent);
+  } else {
+    $agent = inet_aton($agent);
+  }
+  push @enoid, toOID(($ent));
+  push @enoid, encode_ip_address($agent);
+  push @enoid, encode_int($gen);
+  push @enoid, encode_int($spec);
+  push @enoid, encode_timeticks((time-$agent_start_time) * 100);
+  while(@vars) {
+    ($oid) = toOID((shift @vars));
+    $type  = shift @vars;
+    $value = shift @vars;
+    if ($type =~ /string/i) {
+      $value = encode_string($value);
+      push @enoid, [$oid,$value];
+    } elsif ($type =~ /ipaddr/i) {
+      $value = encode_ip_address($value);
+      push @enoid, [$oid,$value];
+    } elsif ($type =~ /int/i) {
+      $value = encode_int($value);
+      push @enoid, [$oid,$value];
+    } elsif ($type =~ /oid/i) {
+      my $tmp = encode_oid_with_errmsg($value);
+      return undef unless defined $tmp;
+      push @enoid, [$oid,$tmp];
+    } else {
+      carp "unknown SNMP type: $type\n"
+	unless ($SNMP_Session::suppress_warnings > 1);
+      return undef;
     }
-    else
-    {
-	$agent = inet_aton($agent);
-    }
-    push @enoid, toOID(($ent));
-    push @enoid, encode_ip_address($agent);
-    push @enoid, encode_int($gen);
-    push @enoid, encode_int($spec);
-    push @enoid, encode_timeticks((time-$agent_start_time) * 100);
-    while(@vars)
-    {
-	($oid) = toOID((shift @vars));
-	$type  = shift @vars;
-	$value = shift @vars;
-	if ($type =~ /string/i)
-	{
-	    $value = encode_string($value);
-	    push @enoid, [$oid,$value];
-	}
-	elsif ($type =~ /int/i)
-	{
-	    $value = encode_int($value);
-	    push @enoid, [$oid,$value];
-	}
-	elsif ($type =~ /oid/i)
-	{
-	    my $tmp = encode_oid_with_errmsg($value);
-	    return undef unless defined $tmp;
-	    push @enoid, [$oid,$tmp];
-	}
-	else
-	{
-	    warn "unknown SNMP type: $type\n"
-		unless ($SNMP_Session::suppress_warnings > 1);
-	    return undef;
-	}
-    }
-    return($session->trap_request_send(@enoid));
+  }
+  return($session->trap_request_send(@enoid));
 }
 
+#
+# A restricted snmpgetbulk.
+#
+sub snmpgetbulk ($$$@) {
+  my($host, $nr, $mr, @vars) = @_;
+  my(@enoid, $var, $response, $bindings, $binding);
+  my($value, $upoid, $oid, @retvals);
+  my($noid);
+  my $session;
+
+  $session = &snmpopen($host, 0, \@vars);
+  if (!defined($session)) {
+    carp "SNMPGETBULK Problem for $host\n"
+      unless ($SNMP_Session::suppress_warnings > 1);
+    return undef;
+  }
+
+  @enoid = &toOID(@vars);
+  return undef unless defined $enoid[0];
+
+  undef @vars;
+  undef @retvals;
+  foreach $noid (@enoid) {
+    $upoid = pretty_print($noid);
+    push(@vars, $upoid);
+  }
+  if ($session->getbulk_request_response($nr, $mr, @enoid)) {
+    $response = $session->pdu_buffer;
+    ($bindings) = $session->decode_get_response($response);
+    while ($bindings) {
+      ($binding, $bindings) = decode_sequence($bindings);
+      ($oid, $value) = decode_by_template($binding, "%O%@");
+      my $tempo = pretty_print($oid);
+      my $tempv = pretty_print($value);
+      push @retvals, "$tempo:$tempv";
+    }
+    return (@retvals);
+  } else {
+    $var = join(' ', @vars);
+    carp "SNMPGETBULK Problem for $var on $host\n"
+      unless ($SNMP_Session::suppress_warnings > 1);
+    return undef;
+  }
+}
+
+#
+# walk a table, calling a user-supplied function for each
+# column of a table.
+#
+sub snmpmaptable($$@) {
+  my($host, $fun, @vars) = @_;
+  my(@enoid, $var, $session);
+
+  $session = &snmpopen($host, 0, \@vars);
+  if (!defined($session)) {
+    carp "SNMPMAPTABLE Problem for $host\n"
+      unless ($SNMP_Session::suppress_warnings > 1);
+    return undef;
+  }
+
+  foreach $var (toOID(@vars)) {
+    push(@enoid, [split('\.', pretty_print($var))]);
+  }
+
+  return $session->map_table([@enoid],
+			      sub() {
+				my ($ind, @vals) = @_;
+				my (@pvals, $val);
+
+				foreach $val (@vals) {
+				  push(@pvals, pretty_print($val));
+				}
+				&$fun($ind, @pvals);
+			      }
+			      );
+}
+
+
 #
 #  Given an OID in either ASN.1 or mixed text/ASN.1 notation, return an
 #  encoded OID.
 #
+sub toOID(@) {
+  my(@vars) = @_;
+  my($oid, $var, $tmp, $tmpv, @retvar);
 
-sub toOID(@)
-{
-    my(@vars) = @_;
-    my($oid, $var, $tmp, $tmpv, @retvar);
+  undef @retvar;
+  foreach $var (@vars) {
+    ($oid, $tmp) = &Check_OID($var);
+    if (!$oid && $SNMP_util::CacheLoaded == 0) {
+      $tmp = $SNMP_Session::suppress_warnings;
+      $SNMP_Session::suppress_warnings = 1000;
 
-    undef @retvar;
-    foreach $var (@vars)
-    {
-	($oid, $tmp) = &Check_OID($var);
-	if (!$oid && $SNMP_util::CacheLoaded == 0)
-	{
-	    $tmp = $SNMP_Session::suppress_warnings;
-	    $SNMP_Session::suppress_warnings = 1000;
+      &snmpLoad_OID_Cache($SNMP_util::CacheFile);
 
-	    &snmpLoad_OID_Cache($SNMP_util::CacheFile);
+      $SNMP_util::CacheLoaded = 1;
+      $SNMP_Session::suppress_warnings = $tmp;
 
-	    $SNMP_util::CacheLoaded = 1;
-	    $SNMP_Session::suppress_warnings = $tmp;
-
-	    ($oid, $tmp) = &Check_OID($var);
-	}
-	while (!$oid && $#SNMP_util::MIB_Files >= 0)
-	{
-	    $tmp = $SNMP_Session::suppress_warnings;
-	    $SNMP_Session::suppress_warnings = 1000;
-
-	    snmpMIB_to_OID(shift(@SNMP_util::MIB_Files));
-
-	    $SNMP_Session::suppress_warnings = $tmp;
-
-	    ($oid, $tmp) = &Check_OID($var);
-	    if ($oid)
-	    {
-		open(CACHE, ">>$SNMP_util::CacheFile");
-		print CACHE "$tmp\t$oid\n";
-		close(CACHE);
-	    }
-	}
-	if ($oid) {
-	    $var =~ s/^$tmp/$oid/;
-	} else {
-	    warn "Unknown SNMP var $var\n"
-		unless ($SNMP_Session::suppress_warnings > 1);
-	    next;
-	}
-	print "toOID: $var\n" if $SNMP_util::Debug;
-	$tmp = encode_oid_with_errmsg($var);
-	return undef unless defined $tmp;
-	push(@retvar, $tmp);
+      ($oid, $tmp) = &Check_OID($var);
     }
-    return @retvar;
+    while (!$oid && $#SNMP_util::MIB_Files >= 0) {
+      $tmp = $SNMP_Session::suppress_warnings;
+      $SNMP_Session::suppress_warnings = 1000;
+
+      snmpMIB_to_OID(shift(@SNMP_util::MIB_Files));
+
+      $SNMP_Session::suppress_warnings = $tmp;
+
+      ($oid, $tmp) = &Check_OID($var);
+      if ($oid) {
+	open(CACHE, ">>$SNMP_util::CacheFile");
+	print CACHE "$tmp\t$oid\n";
+	close(CACHE);
+      }
+    }
+    if ($oid) {
+      $var =~ s/^$tmp/$oid/;
+    } else {
+      carp "Unknown SNMP var $var\n"
+      unless ($SNMP_Session::suppress_warnings > 1);
+      next;
+    }
+    print "toOID: $var\n" if $SNMP_util::Debug;
+    $tmp = encode_oid_with_errmsg($var);
+    return undef unless defined $tmp;
+    push(@retvar, $tmp);
+  }
+  return @retvar;
 }
 
 #
@@ -690,22 +871,23 @@ sub toOID(@)
 #
 sub snmpmapOID(@)
 {
-    my(@vars) = @_;
-    my($oid, $txt, $ind);
+  my(@vars) = @_;
+  my($oid, $txt, $ind);
 
-    $ind = 0;
-    while($ind <= $#vars)
-    {
-	$txt = $vars[$ind++];
-	next unless($txt =~ /^(([a-z][a-z\d\-]*\.)*([a-z][a-z\d\-]*))$/i);
+  $ind = 0;
+  while($ind <= $#vars) {
+    $txt = $vars[$ind++];
+    next unless($txt =~ /^(([a-z][a-z\d\-]*\.)*([a-z][a-z\d\-]*))$/i);
 
-	$oid = $vars[$ind++];
-	next unless($oid =~ /^((\d+.)*\d+)$/);
+    $oid = $vars[$ind++];
+    next unless($oid =~ /^((\d+.)*\d+)$/);
 
-	$SNMP_util::OIDS{$txt} = $oid;
-	print "snmpmapOID: $txt => $oid\n" if $SNMP_util::Debug;
-    }
-    return undef;
+    $SNMP_util::OIDS{$txt} = $oid;
+    $RevNeeded = 1;
+    print "snmpmapOID: $txt => $oid\n" if $SNMP_util::Debug;
+  }
+  
+  return undef;
 }
 
 #
@@ -716,31 +898,28 @@ sub snmpmapOID(@)
 #
 # blank lines and anything after a '#' or between '--' is ignored.
 #
-sub snmpLoad_OID_Cache ($)
-{
-    my($arg) = @_;
-    my($txt, $oid);
+sub snmpLoad_OID_Cache ($) {
+  my($arg) = @_;
+  my($txt, $oid);
 
-    if (!open(CACHE, $arg))
-    {
-	warn "snmpLoad_OID_Cache: Can't open $arg: $!"
-	    unless ($SNMP_Session::suppress_warnings > 1);
-	return -1;
-    }
+  if (!open(CACHE, $arg)) {
+    carp "snmpLoad_OID_Cache: Can't open $arg: $!"
+      unless ($SNMP_Session::suppress_warnings > 1);
+    return -1;
+  }
 
-    while(<CACHE>)
-    {
-	s/#.*//;
-	s/--.*--//g;
-	s/--.*//;
-	next if (/^$/);
-	next unless (/\s/);
-	chop;
-	($txt, $oid) = split(' ', $_, 2);
-	&snmpmapOID($txt, $oid);
-    }
-    close(CACHE);
-    return 0;
+  while(<CACHE>) {
+    s/#.*//;
+    s/--.*--//g;
+    s/--.*//;
+    next if (/^$/);
+    next unless (/\s/);
+    chop;
+    ($txt, $oid) = split(' ', $_, 2);
+    &snmpmapOID($txt, $oid);
+  }
+  close(CACHE);
+  return 0;
 }
 
 #
@@ -748,27 +927,26 @@ sub snmpLoad_OID_Cache ($)
 # Returns the OID and the corresponding text as two separate
 # elements.
 #
-sub Check_OID ($)
-{
-    my($var) = @_;
-    my($tmp, $tmpv, $oid);
+sub Check_OID ($) {
+  my($var) = @_;
+  my($tmp, $tmpv, $oid);
 
-    if ($var =~ /^(([a-z][a-z\d\-]*\.)*([a-z][a-z\d\-]*))/i)
-    {
-	$tmp = $&;
-	$tmpv = $tmp;
-	for (;;) {
-	    last if defined($SNMP_util::OIDS{$tmpv});
-	    last if !($tmpv =~ s/^[^\.]*\.//);
-	}
-	$oid = $SNMP_util::OIDS{$tmpv};
-	if ($oid) {
-	    return ($oid, $tmp);
-	} else {
-	    return undef;
-	}
+  if ($var =~ /^(([a-z][a-z\d\-]*\.)*([a-z][a-z\d\-]*))/i)
+  {
+    $tmp = $&;
+    $tmpv = $tmp;
+    for (;;) {
+      last if defined($SNMP_util::OIDS{$tmpv});
+      last if !($tmpv =~ s/^[^\.]*\.//);
     }
-    return ($var, $var);
+    $oid = $SNMP_util::OIDS{$tmpv};
+    if ($oid) {
+      return ($oid, $tmp);
+    } else {
+      return undef;
+    }
+  }
+  return ($var, $var);
 }
 
 #
@@ -776,28 +954,61 @@ sub Check_OID ($)
 # found in the existing table.  At that time the MIB file will
 # be loaded, and the lookup attempted again.
 #
-sub snmpQueue_MIB_File (@)
-{
-    my(@files) = @_;
-    my($file);
+sub snmpQueue_MIB_File (@) {
+  my(@files) = @_;
+  my($file);
 
-    foreach $file (@files)
-    {
-	push(@SNMP_util::MIB_Files, $file);
-    }
+  foreach $file (@files) {
+    push(@SNMP_util::MIB_Files, $file);
+  }
 }
 
-
 #
 # Read in the passed list of MIB files, parsing them
 # for their text-to-OID mappings
 #
-sub snmpMIB_to_OID ($)
-{
-    my($arg) = @_;
-    my($quote, $buf, $var, $code, $val, $tmp, $tmpv, $strt);
-    my($ret);
-    my(%Link) = (
+sub snmpMIB_to_OID ($) {
+  my($arg) = @_;
+  my($quote, $buf, $var, $code, $val, $tmp, $tmpv, $strt);
+  my($ret);
+  my(%Link) = (
+    'org' => 'iso',
+    'dod' => 'org',
+    'internet' => 'dod',
+    'directory' => 'internet',
+    'mgmt' => 'internet',
+    'mib-2' => 'mgmt',
+    'experimental' => 'internet',
+    'private' => 'internet',
+    'enterprises' => 'private',
+  );
+
+  if (!open(MIB, $arg)) {
+    carp "snmpMIB_to_OID: Can't open $arg: $!"
+      unless ($SNMP_Session::suppress_warnings > 1);
+    return -1;
+  }
+  print "snmpMIB_to_OID: loading $arg\n" if $SNMP_util::Debug;
+  $ret = 0;
+  while(<MIB>) {
+    s/--.*--//g;	# throw away comments (-- anything --)
+    s/--.*//;		# throw away comments (-- anything EOL)
+    if ($quote) {
+      next unless /"/;
+      $quote = 0;
+    }
+    chop;
+#
+#	$buf = "$buf $_";
+# Previous line removed (and following replacement)
+# suggested by Brian Reichert, reichert@numachi.com
+#
+    $buf .= ' ' . $_;
+    $buf =~ s/\s+/ /g;
+
+    if ($buf =~ / DEFINITIONS ::= BEGIN/) {
+      undef %Link;
+      %Link = (
 	'org' => 'iso',
 	'dod' => 'org',
 	'internet' => 'dod',
@@ -807,143 +1018,89 @@ sub snmpMIB_to_OID ($)
 	'experimental' => 'internet',
 	'private' => 'internet',
 	'enterprises' => 'private',
-    );
+      );
+    }
 
+    $buf =~ s/OBJECT-TYPE/OBJECT IDENTIFIER/;
+    $buf =~ s/OBJECT-IDENTITY/OBJECT IDENTIFIER/;
+    $buf =~ s/MODULE-IDENTITY/OBJECT IDENTIFIER/;
+    $buf =~ s/ IMPORTS .*\;//;
+    $buf =~ s/ SEQUENCE {.*}//;
+    $buf =~ s/ SYNTAX .*//;
+    $buf =~ s/ [\w-]+ ::= OBJECT IDENTIFIER//;
+    $buf =~ s/ OBJECT IDENTIFIER .* ::= {/ OBJECT IDENTIFIER ::= {/;
+    $buf =~ s/".*"//;
+    if ($buf =~ /"/) {
+      $quote = 1;
+    }
 
-    if (!open(MIB, $arg))
-    {
-	warn "snmpMIB_to_OID: Can't open $arg: $!"
+    if ($buf =~ / ([\w\-]+) OBJECT IDENTIFIER ::= {([^}]+)}/) {
+      $var = $1;
+      $buf = $2;
+      undef $val;
+      $buf =~ s/ +$//;
+      ($code, $val) = split(' ', $buf, 2);
+
+      if (!defined($val) || (length($val) <= 0)) {
+	$SNMP_util::OIDS{$var} = $code;
+	$ret++;
+	print "'$var' => '$code'\n" if $SNMP_util::Debug;
+      } else {
+	$strt = $code;
+	while($val =~ / /) {
+	  ($tmp, $val) = split(' ', $val, 2);
+	  if ($tmp =~ /([\w\-]+)\((\d+)\)/) {
+	    $tmp = $1;
+	    $tmpv = "$SNMP_util::OIDS{$strt}.$2";
+	    $Link{$tmp} = $strt;
+	    if (defined($SNMP_util::OIDS{$tmp})) {
+	      if ($tmpv ne $SNMP_util::OIDS{$tmp}) {
+		$strt = "$strt.$tmp";
+		$SNMP_util::OIDS{$strt} = $tmpv;
+		$ret++;
+	      }
+	    } else {
+	      $SNMP_util::OIDS{$tmp} = $tmpv;
+	      $ret++;
+	      $strt = $tmp;
+	    }
+	  }
+	}
+
+	if (!defined($SNMP_util::OIDS{$strt})) {
+	  carp "snmpMIB_to_OID: $arg: \"$strt\" prefix unknown, load the parent MIB first.\n"
 	    unless ($SNMP_Session::suppress_warnings > 1);
-	return -1;
+	}
+	$Link{$var} = $strt;
+	$val = "$SNMP_util::OIDS{$strt}.$val";
+	if (defined($SNMP_util::OIDS{$var})) {
+	  if ($val ne $SNMP_util::OIDS{$var}) {
+	    $var = "$strt.$var";
+	  }
+	}
+
+	$SNMP_util::OIDS{$var} = $val;
+	$ret++;
+
+	print "'$var' => '$val'\n" if $SNMP_util::Debug;
+      }
+      undef $buf;
     }
-    print "snmpMIB_to_OID: loading $arg\n" if $SNMP_util::Debug;
-    $ret = 0;
-    while(<MIB>)
-    {
-	s/--.*--//g;		# throw away comments (-- anything --)
-	s/--.*//;		# throw away comments (-- anything EOL)
-	if ($quote)
-	{
-	    next unless /"/;
-	    $quote = 0;
-	}
-	chop;
-#
-#	$buf = "$buf $_";
-# Previous line removed (and following replacement)
-# suggested by Brian Reichert, reichert@numachi.com
-#
-	$buf .= ' ' . $_;
-	$buf =~ s/\s+/ /g;
-
-	if ($buf =~ / DEFINITIONS ::= BEGIN/)
-	{
-	    undef %Link;
-	    %Link = (
-		'org' => 'iso',
-		'dod' => 'org',
-		'internet' => 'dod',
-		'directory' => 'internet',
-		'mgmt' => 'internet',
-		'mib-2' => 'mgmt',
-		'experimental' => 'internet',
-		'private' => 'internet',
-		'enterprises' => 'private',
-	    );
-	}
-
-	$buf =~ s/OBJECT-TYPE/OBJECT IDENTIFIER/;
-	$buf =~ s/MODULE-IDENTITY/OBJECT IDENTIFIER/;
-	$buf =~ s/ IMPORTS .*\;//;
-	$buf =~ s/ SEQUENCE {.*}//;
-	$buf =~ s/ SYNTAX .*//;
-	$buf =~ s/ [\w-]+ ::= OBJECT IDENTIFIER//;
-	$buf =~ s/ OBJECT IDENTIFIER .* ::= {/ OBJECT IDENTIFIER ::= {/;
-	$buf =~ s/".*"//;
-	if ($buf =~ /"/)
-	{
-	    $quote = 1;
-	}
-
-	if ($buf =~ / ([\w\-]+) OBJECT IDENTIFIER ::= {([^}]+)}/) {
-	    $var = $1;
-	    $buf = $2;
-	    undef $val;
-	    $buf =~ s/ +$//;
-	    ($code, $val) = split(' ', $buf, 2);
-
-	    if (length($val) <= 0)
-	    {
-		$SNMP_util::OIDS{$var} = $code;
-		$ret++;
-		print "'$var' => '$code'\n" if $SNMP_util::Debug;
-	    }
-	    else
-	    {
-		$strt = $code;
-
-		while($val =~ / /)
-		{
-		    ($tmp, $val) = split(' ', $val, 2);
-		    if ($tmp =~ /([\w\-]+)\((\d+)\)/)
-		    {
-			$tmp = $1;
-			$tmpv = "$SNMP_util::OIDS{$strt}.$2";
-			$Link{$tmp} = $strt;
-			if (defined($SNMP_util::OIDS{$tmp}))
-			{
-			    if ($tmpv ne $SNMP_util::OIDS{$tmp})
-			    {
-				$strt = "$strt.$tmp";
-				$SNMP_util::OIDS{$strt} = $tmpv;
-				$ret++;
-			    }
-			}
-			else
-			{
-			    $SNMP_util::OIDS{$tmp} = $tmpv;
-			    $ret++;
-			    $strt = $tmp;
-			}
-		    }
-		}
-
-		if (!defined($SNMP_util::OIDS{$strt}))
-		{
-		    warn "snmpMIB_to_OID: $arg: \"$strt\" prefix unknown, load the parent MIB first.\n"
-			unless ($SNMP_Session::suppress_warnings > 1);
-		}
-		$Link{$var} = $strt;
-		$val = "$SNMP_util::OIDS{$strt}.$val";
-		if (defined($SNMP_util::OIDS{$var}))
-		{
-		    if ($val ne $SNMP_util::OIDS{$var})
-		    {
-			$var = "$strt.$var";
-		    }
-		}
-
-		$SNMP_util::OIDS{$var} = $val;
-		$ret++;
-
-		print "'$var' => '$val'\n" if $SNMP_util::Debug;
-	    }
-	    undef $buf;
-	}
-    }
-    close(MIB);
-    return $ret;
+  }
+  close(MIB);
+  $RevNeeded = 1;
+  return $ret;
 }
 
 sub encode_oid_with_errmsg ($) {
-    my ($oid) = @_;
-    my $tmp = encode_oid(split(/\./, $oid));
-    if (! defined $tmp) {
-	warn "cannot encode Object ID $oid: $BER::errmsg"
-	    unless ($SNMP_Session::suppress_warnings > 1);
-	return undef;
-    }
-    return $tmp;
+  my ($oid) = @_;
+  my $tmp = encode_oid(split(/\./, $oid));
+  if (! defined $tmp) {
+    carp "cannot encode Object ID $oid: $BER::errmsg"
+      unless ($SNMP_Session::suppress_warnings > 1);
+    return undef;
+  }
+  return $tmp;
 }
 
 1;
